"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var t = __importStar(require("@babel/types"));
var plugin_syntax_jsx_1 = __importDefault(require("@babel/plugin-syntax-jsx"));
var KnownAttrs = [
    "class",
    "staticClass",
    "style",
    "key",
    "ref",
    "refInFor",
    "slot",
    "scopedSlots"
];
var ModifierFuncNames = {
    sync: "__sync",
    relay: "__relay",
    once: "__once",
    capture: "__capture",
    passive: "__passive",
    captureOnce: "__captureOnce"
};
var EventPrefixes = {
    once: "~",
    capture: "!",
    passive: "&",
    captureOnce: "~!"
};
function toCamelCase(value) {
    return value.replace(/\-([a-z])/g, function (_, ch) { return ch.toUpperCase(); });
}
function getUnusedVariableName(scope, prefix) {
    var bindings = scope.bindings;
    for (var i = 0;; ++i) {
        var varName = prefix + i.toString();
        if (!bindings[varName]) {
            return varName;
        }
    }
}
function getJSXAttributeName(node) {
    var name = node.name;
    if (t.isJSXIdentifier(name)) {
        return name.name;
    }
    else {
        // name is JSXNamespacedName
        return name.namespace.name + ":" + name.name.name;
    }
}
function getModifierName(callee) {
    var name;
    if (t.isIdentifier(callee)) {
        name = callee.name;
    }
    else if (t.isMemberExpression(callee)) {
        var object = callee.object, property = callee.property;
        if (t.isIdentifier(object) && t.isIdentifier(property)) {
            name = property.name;
        }
        else {
            return undefined;
        }
    }
    else {
        return undefined;
    }
    for (var key in ModifierFuncNames) {
        var modifier = key;
        if (name === ModifierFuncNames[modifier]) {
            return modifier;
        }
    }
    return undefined;
}
function createAssignmentFunction(target, argName) {
    /*
     * <argName> => {
     *   <target> = <argName>;
     * }
     */
    return t.arrowFunctionExpression([t.identifier(argName)], t.blockStatement([
        t.expressionStatement(t.assignmentExpression("=", target, t.identifier(argName)))
    ]));
}
function createEmitFunction(eventName, argName) {
    /*
     * <argName> => {
     *   this.$emit("update:<propName>", <argName>);
     * }
     */
    return t.arrowFunctionExpression([t.identifier(argName)], t.blockStatement([
        t.expressionStatement(t.callExpression(t.memberExpression(t.thisExpression(), t.identifier("$emit")), [eventName, t.identifier(argName)]))
    ]));
}
function processSyncModifier(state, modifierName, path, jsxAttrPath) {
    /**
     *  <tag propName={ __sync(foo.bar) }>...
     *        \         \_path  \_content
     *         \_jsxAttr
     */
    var attrName = getJSXAttributeName(jsxAttrPath.node);
    if (KnownAttrs.indexOf(attrName) >= 0 ||
        /^(domProps|on|nativeOn|hook)[\-_A-Z]/.test(attrName)) {
        throw path.buildCodeFrameError(modifierName + " modifier can be used only in component prop");
    }
    if (path.node.arguments.length !== 1) {
        throw path.buildCodeFrameError(modifierName + " modifier must have one argument");
    }
    var arg = path.node.arguments[0];
    if (!t.isMemberExpression(arg)) {
        throw path.buildCodeFrameError("argument of " + modifierName + " modifier must be MemberExpression");
    }
    if (modifierName == "sync") {
        // remove sync call
        path.replaceWith(arg);
        // add update handler
        state.on.push(t.objectProperty(t.stringLiteral("update:" + toCamelCase(attrName)), createAssignmentFunction(arg, getUnusedVariableName(path.scope, "_v"))));
    }
    else {
        // relay
        var eventName = void 0;
        if (arg.computed) {
            if (t.isStringLiteral(arg.property)) {
                eventName = t.stringLiteral("update:" + arg.property.value);
            }
            else {
                eventName = t.binaryExpression("+", t.stringLiteral("update:"), arg.property);
            }
        }
        else if (t.isIdentifier(arg.property)) {
            eventName = t.stringLiteral("update:" + arg.property.name);
        }
        else {
            throw path.buildCodeFrameError("Failed to determine event name to emit");
        }
        // remove sync call
        path.replaceWith(arg);
        // add update handler
        state.on.push(t.objectProperty(t.stringLiteral("update:" + toCamelCase(attrName)), createEmitFunction(eventName, getUnusedVariableName(path.scope, "_v"))));
    }
}
function processEventModifier(state, modifierName, path, jsxAttrPath) {
    /**
     *  <tag onClick={ __capture(handler) }>...
     *        \        \_path    \_content
     *         \_jsxAttr
     */
    var attrName = getJSXAttributeName(jsxAttrPath.node);
    var match = /^(on|nativeOn)([\-_A-Z])(.*)$/.exec(attrName);
    if (!match) {
        throw path.buildCodeFrameError(modifierName + " modifier can be used only in event handler");
    }
    var eventType = match[1];
    var eventName = (match[2] === "-" ? "" : match[2].toLowerCase()) + match[3];
    var prefix = EventPrefixes[modifierName];
    if (path.node.arguments.length !== 1) {
        throw path.buildCodeFrameError(modifierName + " modifier must have one argument");
    }
    var arg = path.node.arguments[0];
    if (t.isSpreadElement(arg)) {
        throw path.buildCodeFrameError(modifierName + " modifier have one argument");
    }
    if (t.isJSXNamespacedName(arg)) {
        throw path.buildCodeFrameError(modifierName + " modifier argument must not be JSXNamespacedName");
    }
    if (t.isArgumentPlaceholder(arg)) {
        throw path.buildCodeFrameError(modifierName + " modifier argument must not be ArgumentPlaceholder");
    }
    // remove JSXAttribute
    jsxAttrPath.remove();
    // add prefixed event handler
    (eventType === "on" ? state.on : state.nativeOn).push(t.objectProperty(t.stringLiteral(prefix + eventName), arg));
}
var jsxOpeningElementVisitor = {
    CallExpression: function (path) {
        /**
         * Transform modifier calls
         *
         * Code pattern to be transformed:
         *  <tag propName={ __sync(foo.bar) }>...
         *        \        \ \__target [CallExpression]
         *         \        \__parent [JSXExpressionContainer]
         *          \__grand parent [JSXAttribute]
         */
        if (!t.isJSXExpressionContainer(path.parent)) {
            return;
        }
        var grandParentPath = path.parentPath.parentPath;
        if (!t.isJSXAttribute(grandParentPath.node)) {
            return;
        }
        var jsxAttrPath = grandParentPath;
        var modifier = getModifierName(path.node.callee);
        if (!modifier) {
            return;
        }
        if (modifier === "sync" || modifier === "relay") {
            processSyncModifier(this, modifier, path, jsxAttrPath);
        }
        else {
            processEventModifier(this, modifier, path, jsxAttrPath);
        }
    }
};
module.exports = function () {
    var plugin = {
        name: "vue-jsx-modifier",
        inherits: plugin_syntax_jsx_1.default,
        visitor: {
            Program: function (path) {
                path.traverse({
                    JSXOpeningElement: function (path) {
                        var on = [];
                        var nativeOn = [];
                        path.traverse(jsxOpeningElementVisitor, { on: on, nativeOn: nativeOn });
                        if (on.length > 0 || nativeOn.length > 0) {
                            var vnodeData = t.objectExpression([]);
                            for (var _i = 0, _a = ["on", "nativeOn"]; _i < _a.length; _i++) {
                                var key = _a[_i];
                                var obj = key === "on" ? on : nativeOn;
                                if (obj.length == 0) {
                                    continue;
                                }
                                vnodeData.properties.push(t.objectProperty(t.identifier(key), t.objectExpression(obj)));
                            }
                            path.node.attributes.push(t.jsxSpreadAttribute(vnodeData));
                        }
                    }
                });
            }
        }
    };
    return plugin;
};
//# sourceMappingURL=index.js.map